%
% Coded by Rami Mowakeaa (ramo1@umbc.edu) based on earlier IVA work by Matt
% Anderson
%
% References:
%
% Mowakeaa, R., Boukouvalas, Z., Adal?, T., & Cavalcante, C. (2016, July). 
%    On the characterization, generation, and efficient estimation of the 
%    complex multivariate GGD. In 2016 IEEE Sensor Array and Multichannel 
%    Signal Processing Workshop (SAM) (pp. 1-5). IEEE.
%
% Mowakeaa, R., Boukouvalas, Z., Long, Q., & Adali, T. (2019). IVA using 
%    complex multivariate GGD: application to fMRI analysis. 
%    Multidimensional Systems and Signal Processing, 1-20.
%
function [s,p] = CompareIVAoverT(p)
%to compare different IVA methods: GL, MG, CMGGD for paper
%input p is a structure with problem parameters
%output s is a structure with all the resutls

rng(1)

if ~exist('p','var') %setup problem and global constants
    p = ProbSetup();
else
    p = ProbSetup(p);
end
s.isi = zeros(4,length(p.TT),p.runs);
s.stats = zeros(p.runs,1);
s.iters = zeros(3,length(p.TT),p.runs);

tic
for t = 1:length(p.TT)
    p = CircSetup(p,1); %setup variables that change with each doi
    p.T = p.TT(t);
    for r = 1:p.runs
        p = RunSetup(p); %setup variables that change with each run
        fprintf('\nT= %d, RUN %d\n',p.T,r);
        
        %run IVA methods
        %
        Wt = iva_second_order(p.X, 'W_init', p.Winit);
        %p.Winit = Wt;
        
		res1 = iva_cmggd(p.X, 0, 'whiten', true, 'initW', p.Winit, 'verbose', false);
		p.W(:,:,:,1) = res1.W;
		s.iters(1,t,r) = res1.iter;
        
        res2 = iva_cmggd(p.X, 1, 'whiten', true, 'initW', p.Winit, 'verbose', false);
		p.W(:,:,:,2) = res2.W;
		s.iters(2,t,r) = res2.iter;
        
        res3 = iva_cmggd(p.X, 2, 'whiten', true, 'initW', p.Winit, 'verbose', false);
        p.W(:,:,:,3) = res3.W;
		s.iters(3,t,r) = res3.iter;
		
        if res1.status == 3 || res2.status == 3 || res3.status == 3
            s.stats(r) = 1;
        end
        
        p.W(:,:,:,4) = iva_laplace(p.X, 'initW', Wt);
        p.W(:,:,:,5) = Wt;
        %
        for m = 1:5
            [~,s.isi(m,t,r),~] = bss_isi(p.W(:,:,:,m),p.A);
        end
    end
    %str = sprintf('t%d.mat',p.T);
    %save(str,'s');
end

s = PostProcess(p,s);

fprintf('Total run time %.2f mins\n',toc/60);

end

%% **********************************************************************
%
%
%
%
function p = ProbSetup(p)
%setup any missing parameters
if ~exist('p','var')
    %     p.TT = 10000;
    p.TT = 5000:5000:25000;
    %     p.TT = 1000:2000:19000;
    %     p.T = 10.^(3:5);%10000; %number of samples
    p.N = 3; %number of sources
    p.K = 4; %number of datasets
    
    p.runs = 10; %number of runs
    
    p.Crho = .9*rand(p.N,1); %rho for structured C generation
    
    p.beta_high = 8; %max source shape parameter
    p.beta_low = .125; %min source shape parameter
    
    %for noncircularity, we need a single metric from which to derive the
    %noncircularity coefficients of each of the marginals
    %we opt for \rho3 (eq 3.54 in Schreier, et al) which is average sum of
    %squares of noncircularity coefficients
    p.doi = .5;%0:.45:.9; %degree-of-impropriety (overall circularity coefficient)
    
    p.lw = 3; %LineWidth plot parameter
    p.ms = 12; %MarkerSize plot parameter
    p.fs = 18; %FontSize plot parameter
else
    if ~isfield(p,'T')
        p.T = 10000;
    end
    if ~isfield(p,'N')
        p.N = 2;
    end
    if ~isfield(p,'K')
        p.K = 2;
    end
    if ~isfield(p,'runs')
        p.runs = 2;
    end
    if ~isfield(p,'beta_high')
        p.beta_high = .5;
    end
    if ~isfield(p,'beta_low')
        p.beta_low = .125;
    end
    if ~isfield(p,'doi')
        p.doi = .1:.4:.9;
    end
    if ~isfield(p, 'lw')
        p.lw = 3;
    end
    if ~isfield(p, 'ms')
        p.ms = 12;
    end
    if ~isfield(p, 'fs')
        p.fs = 18;
    end
end

end

function p = RunSetup(p)
%compute run-dependent vars
% shape parameters p.betan
% augmented covariance matrices p.Ca
% W initialization p.Winit
% CMGGD sources p.S
% linear mixtures p.X
% mixing matrices p.A

p.betan = p.beta_low + (p.beta_high-p.beta_low).*rand(p.N,1); %shape parameters



p.Ca = covRandom(p);
%p.Ca = covStructured2(p);

Winit = randn(p.N,p.N,p.K) + 1j*randn(p.N,p.N,p.K);
for k = 1:p.K
    for n = 1:p.N
        Winit(n,:,k) = Winit(n,:,k) ./ norm(Winit(n,:,k),2);
    end
end
p.Winit = Winit;
p.Winit = repmat(eye(p.N),1,1,p.K);

S = zeros(p.N,p.T,p.K); %CMGGD sources
for n = 1:p.N
    Cr = MatrixAugmentedToBivariate(p.Ca(:,:,n));
    Sb = generate_MGGD(p.T, 2*p.K, Cr, p.betan(n))';
    S(n,:,:) = Sb(:,1:p.K) + 1j*Sb(:,p.K+1:end);
end
p.S = S;

A = rand(p.N,p.N,p.K) + 1j*rand(p.N,p.N,p.K); %mixing matrices
for k = 1:p.K
    for n = 1:p.N
        A(n,:,k) = A(n,:,k) ./ norm(A(n,:,k),2);
    end
end
p.A = A;

X = zeros(p.N,p.T,p.K); %linear mixtures
for k = 1:p.K
    X(:,:,k) = A(:,:,k) * S(:,:,k);
end
p.X = X;

p.W = zeros(p.N,p.N,p.K,5);

end

function Ca = covRandom(p)
Ca = zeros(2*p.K,2*p.K,p.N); %augmented covariance matrices
for n=1:p.N
    Un = (randn(p.K) + 1j*randn(p.K));
    for k=1:p.K
        Un(:,k)=Un(:,k)./norm(Un(:,k));
    end
    C = Un' * Un; %Hermitian, positive definite
    
    if any(p.nc) %~= 0
        Csqr = sqrtm(C);
        F = UnitaryM(p.K);
        Cp = Csqr * F * diag(p.nc(n,:)) * F.' * Csqr.';
    else
        Cp = zeros(p.K);
    end
    %verification
    % since we only wish to verify the circularity coefficients, a Takagi
    % factorization isn't necessary.  regular SVD will do
    %Cinvsqr = inv(sqrtm(C));
    %CV = Cinvsqr * Cp * Cinvsqr.';
    %[~,K,~] = svd(CV);
    
    Ca(:,:,n) = [C Cp; conj(Cp) conj(C)];
    
    %verification
    % positive definiteness
    %[~,p] = chol(Ca(:,:,n));
    %assert(p==0);
end
end

function p = CircSetup(p,d)

p.nc = sqrt(p.doi(d)) * ones(p.N,p.K); %noncircularity coefficients

end

function [U,verify]= UnitaryM(n)
% generate a random complex matrix
X = complex(rand(n),rand(n))/sqrt(2);
% factorize the matrix
[Q,R] = qr(X);
R = diag(diag(R)./abs(diag(R)));
% unitary matrix
U = Q*R;
% verification
verify = U*U';
end

function s = PostProcess(p,s)
%perform some post processing using results

s.adv = 100 * [s.isi(2,:,:) ./ s.isi(1,:,:); ...
    s.isi(3,:,:) ./ s.isi(1,:,:); ...
    s.isi(4,:,:) ./ s.isi(1,:,:); ...
    s.isi(5,:,:) ./ s.isi(1,:,:)] - 100;

s.advAve = mean(s.adv,3);
s.advStd = std(s.adv,0,3);

s.isiAve = mean(s.isi,3);
s.isiStd = std(s.isi,0,3);

s.iterAve=mean(s.iters,3);
s.iterStd = std(s.iters,0,3);


%mean value absolute figure
ttt=p.TT;
figure(1)
semilogy(ttt,s.isiAve(1,:),'g--<', ttt,s.isiAve(2,:),'bo--', ttt,s.isiAve(3,:),'rx--', ttt,s.isiAve(4,:),'k*--', ttt,s.isiAve(5,:),'c->', ...
    'LineWidth',p.lw);
legend('IVA-CMGGD-SD','IVA-CMGGD-BFGS','IVA-CMGGD-L-BFGS','IVA-GL','IVA-G');
xlabel('Sample Size')
ylabel('jISI')

figure(2)
plot(ttt,s.iterAve(1,:),'g--<', ttt,s.iterAve(2,:),'bo--', ttt,s.iterAve(3,:),'rx--', ...
    'LineWidth',p.lw);
legend('IVA-CMGGD-SD','IVA-CMGGD-BFGS','IVA-CMGGD-L-BFGS');
xlabel('Sample Size')
ylabel('Iterations')
end