%
% Coded by Rami Mowakeaa (ramo1@umbc.edu) based on earlier IVA work by Matt
% Anderson. (see references below) and
% for a general description of the algorithm and its relationship with others,
% see http://mlsp.umbc.edu/jointBSS_introduction.html
%
% References:
%
% Mowakeaa, R., Boukouvalas, Z., Adal?, T., & Cavalcante, C. (2016, July).
%    On the characterization, generation, and efficient estimation of the
%    complex multivariate GGD. In 2016 IEEE Sensor Array and Multichannel
%    Signal Processing Workshop (SAM) (pp. 1-5). IEEE.
%
% Mowakeaa, R., Boukouvalas, Z., Long, Q., & Adali, T. (2019). IVA using
%    complex multivariate GGD: application to fMRI analysis.
%    Multidimensional Systems and Signal Processing, 1-20.
%
function res = iva_cmggd(X,BFGS,varargin)
%[W,cost,isi] = iva_mg_complex_nc(X,varargin)
%
% Options            Values         Description
%
% verbose            true/false     set to true to display iterations results
% maxIter            scalar>0       maximum number of iterations
% supplyA            A              clariyoant demixing matrices, A is NxNxK; (debug)
% initW              W              initial demixing matrices, W is NxNxK
% ver                'grad',        gradient version of iva_mg_complex
%                    'newton',      Newton version of iva_mg_complex
%                    'quasinewton'  Quasi-Newton version of iva_mg_complex
% whiten             true/false     set to true to display iterations results
% eps                scalar>0       minimum change in W to continue iterations
% alpha0             0<scalar<=1    initial step size (adaptively reduced)

verbose=false;
maxIter = 4*512;
supplyA=false;
supplyW=false;
whiten_opt=true;
eps = 1e-14;
gradEps = 5e-3;
sigmaEps = 1e-2;
gradDelta = 1;
alpha0=1e0;
isi=zeros(1,maxIter);

% other parameters
blowup = 1e3;
alphaMin=1e-4;
outputISI=false;
ssv = 2.^(-4:-1:-8);

BM = 3; % number of vectors stored for hessian estimate
if ~exist('BFGS','var')
	BFGS = 2;
end
[N,~,K] = size(X);
Einterval = floor(sqrt(N^2*K));
Eflag = true;
Dflag = false;

% Reset default values above with optional parameters:
nOpts2=length(varargin);
if mod(nOpts2,2)~=0
	error('Option parameter needs to be specified')
end
nOpts=nOpts2/2;
for iOpt=1:nOpts
	param_description=varargin{2*(iOpt-1)+1};
	param_value=varargin{2*(iOpt-1)+2};
	switch lower(param_description)
		case 'verbose'
			verbose=logical(param_value);
		case 'maxiter'
			maxIter=param_value;
		case 'supplya'
			supplyA=true;
			A=param_value;
		case 'initw'
			supplyW=true;
			W=param_value;
		case 'alpha0'
			alpha0=param_value;
		case 'eps'
			eps=param_value;
		case 'normw'
			normW = param_value; % 1 rows are unit-norm
		case 'bm'
			BM = param_value;
		case 'whiten'
			whiten_opt=logical(param_value);
		otherwise
			error('Unknown Parameter Description')
	end % switch param_description
end

if ~exist('normW','var')
	normW = 1;
end

if whiten_opt
	[X,V,~]=whiten(X);
end

if ~supplyW
	for k = 1:K
		W(:,:,k) = eye(N);
	end
end

[N,T,K]=size(X); % the input format insures all datasets have equal number of samples
x_cell=cell(K,1);
for k=1:K
	x_cell{k}=squeeze(X(:,:,k));
end
Rx=cell(K,K);
Px=cell(K,K);
for k1=1:K
	Rx{k1,k1}=1/T*(x_cell{k1}*x_cell{k1}');
	Px{k1,k1}=1/T*(x_cell{k1}*x_cell{k1}.');
	for k2=(k1+1):K
		Rx{k1,k2}=1/T*(x_cell{k1}*x_cell{k2}');
		Rx{k2,k1}=Rx{k1,k2}';
		
		Px{k1,k2}=1/T*(x_cell{k1}*x_cell{k2}.');
		Px{k2,k1}=Px{k1,k2}.';
		
	end
end

if supplyA
	% only reason to supply A matrices is to display running performance
	%    verbose=true;
	
	if nargout>2
		outputISI=true;
		isi=zeros(1,maxIter);
	end
	if input_cells
		A_cell=A;
		A=zeros(N,N,K);
		for k=1:K
			A(:,:,k)=A_cell{k};
		end
	end
	if whiten_opt
		if input_cells
			for k=1:K
				A(:,:,k) = V{k}*A(:,:,k);
			end
		else
			for k=1:K
				A(:,:,k) = V(:,:,k)*A(:,:,k);
			end
		end
	end
end

Rxall=cell2mat(Rx);
r=rank(Rxall);
if r<(N*K)
	% inflate Rx
	eigRx=eigsort(Rxall);
	Rxall=Rxall+eigRx(r)*eye(N*K); %diag(linspace(0.1,0.9,N*K)); % diagonal loading
	Rx=mat2cell(Rxall,N*ones(K,1),N*ones(K,1));
end

for n=1:N
	if normW == 1
		for k=1:K
			W(n,:,k) = W(n,:,k) ./ norm(W(n,:,k),2);
		end
	else
		ww = W(:,:,k);
		W(:,:,k) = ww / sqrtm(ww'*ww);
	end
end

cost=zeros(1,maxIter);
costTrue = cost;

beta_final = zeros(N,1);

Y = zeros(N,T,K);
magGrad = zeros(maxIter,1);
a0 = zeros(maxIter,1);
mdw = zeros(maxIter,1);
gradAve = zeros(maxIter,1);
sigmaChange = zeros(maxIter,1);
Sigma_Total = zeros(2*K,2*K,N);
sigmab = Sigma_Total;
sigmab_old = Sigma_Total;
etan = zeros(N,1);
sslast = zeros(1,3);
Xc = conj(X);
dW = zeros(N,N,K);
if BFGS == 1
	Bk = eye(2*N^2*K);
end
W_old = W;
YM = zeros(N^2*K,BM);
SM = YM;
aa = zeros(BM,1);
rr = zeros(BM,1);

disp(['Performing IVA-CMGGD ', num2str(BFGS)]);

for iter = 1:maxIter
	sslast = [alpha0 sslast(1:2)];
	if mod(iter,1000)==0 || Dflag,disp(['iter ' num2str(iter) ' MagGrad ' num2str(magGrad(iter-1)) ' Cost ' num2str(costmin)]);end
	max_delta_w=0;
	dWp = dW;
	
	if supplyA
		[~,amari_ivag2]=bss_isi(W,A);
		if outputISI
			isi(iter)=amari_ivag2;
		else
			isi(iter) = NaN;
		end
	else
		isi(iter) = NaN;
	end
	
	for k = 1:K
		Y(:,:,k) = (W(:,:,k)) * X(:,:,k);
	end
	
	WnoutM = cell(N,1);
	for nn=1:N
		nout=[1:nn-1 nn+1:N];
		WnoutM{nn} = W(nout,:,:);
	end
	for n=1:N
		Wn = W(n,:,:);
		Wnout = WnoutM{n};
		
		if K == 1
			Yn = Y(n,:);
		else
			Yn = squeeze(Y(n,:,:)).';
		end
		
		Yna = [Yn; conj(Yn)];
		Ynb = [real(Yn); imag(Yn)];
		
		sigmab_old = sigmab;
		if mod(iter,Einterval)==1 || Eflag
			lastSigmaEval = iter;
			Sigma_n=eye(K);
			SigmaP_n = zeros(K);
			for sk1=1:K
				for sk2=sk1:K
					Sigma_n(sk1,sk2)=Wn(1,:,sk1)*Rx{sk1,sk2}*Wn(1,:,sk2)';
					SigmaP_n(sk1,sk2)=Wn(1,:,sk1)*Px{sk1,sk2}*Wn(1,:,sk2).';
					if sk1~=sk2
						Sigma_n(sk2,sk1)=Sigma_n(sk1,sk2)'; % conjugate symmetric
						SigmaP_n(sk2,sk1)=SigmaP_n(sk1,sk2); % real symmetric
					end
				end % k2
			end %k3
			Sigma_Total = [Sigma_n SigmaP_n; SigmaP_n' conj(Sigma_n)];
			sigmab(:,:,n) = MatrixAugmentedToBivariate(Sigma_Total);
			[betan,~] = EstimateBetaML(Ynb,sigmab(:,:,n));
		else
			betan = beta_final(n);
		end
		
		sigmaInvYna = MatrixBivariateToAugmented(sigmab(:,:,n)) \ Yna;
		
		quadd = real(dot(Yna, sigmaInvYna));
		qbm1 = quadd.^(betan-1);
		
		beta_final(n) = betan;
		
		etanln = 1/betan*log(2) + gammaln((K+1)/betan) - gammaln(K/betan) - log(2*K);
		etan(n) = exp(etanln);
		
		dW(n,:,:) = FunGrad(W(n,:,:),Wnout,qbm1,sigmaInvYna,Xc,etan(n),betan);
		
	end %n
	
	%% update hessian estiamte
	if iter > 1
		sk = sk(:);
		yk = dW - dWp;
		yk = yk(:);
		
		if BFGS == 1
			ska = [sk; conj(sk)];
			yka = [yk; conj(yk)];
			
			ykahska = real(yka' * ska);
			
			B11 = sk * yk';
			B12 = sk * yk.';
			B1 = [B11 B12; conj(B12) conj(B11)];
			
			Bk = (eye(2*N^2*K) - B1 / ykahska) * Bk * (eye(2*N^2*K) - B1' / ykahska) + (ska*ska') / ykahska;
			
			dW_aug = [dW(:); conj(dW(:))];
			pk_aug = -Bk * dW_aug;
			pk = reshape(pk_aug(1:N^2*K),N,N,K);
			
		elseif BFGS == 2 % L-BFGS
			
			% L-BFGS
			
			pk = -dW(:);
			YM = [yk YM(:,1:BM-1)];
			SM = [sk SM(:,1:BM-1)];
			BMind = min(iter-1,BM);
			for m = 1:BMind
				rr(m) = real(YM(:,m)' * SM(:,m));
				aa(m) = real(SM(:,m)' * pk) / rr(m);
				pk = pk - aa(m) * YM(:,m);
			end
			
			%             pk = pk * real(yk' * sk) / (yk' * yk) / 2;
			pk = pk * real(YM(:,1)' * SM(:,1)) / (YM(:,1)' * YM(:,1));
			
			for m = 1:BMind
				bb = real(YM(:,m)' * pk) / rr(m);
				pk = pk + (aa(m) - bb) * SM(:,m);
			end
			pk = reshape(pk,N,N,K);
		elseif BFGS == 0 % SD
			pk = -dW;
		end
	else
		pk = -dW;
	end
	
	% project descent direction
	
	pkp = pk;
	for n=1:N
		for k=1:K
			gradProjMat = W(n,:,k).' * conj(W(n,:,k));
			pkp(n,:,k) = ((eye(N) - gradProjMat) * pk(n,:,k).').';
		end
	end
	
	%% compute step size and next W
	% first compute lofAlpha
	ssc = 1e-4; % step size proportion constant c1
	fW = FunCost(W,X,sigmab,beta_final,etan); % function value at W
	dd = real(sum(sum(sum(conj(pkp) .* dW)))); % directional derivative g' * p for SD
	
	magSq = conj(pkp) .* pkp;
	magGrad(iter) = sqrt(sum(magSq(:)));
	
	Wopts = zeros(N,N,K,length(ssv));
	costopts = zeros(length(ssv),1);
	for ssk = 1:length(ssv)
		Wopts(:,:,:,ssk) = FunNorm(W + ssv(ssk) * pkp, normW);
		
		costopts(ssk)= FunCost(Wopts(:,:,:,ssk),X,sigmab,beta_final,etan);
		threshold = fW + ssc * ssv(ssk) * dd;
		
		if costopts(ssk) < threshold
			W = Wopts(:,:,:,ssk);
			costmin = costopts(ssk);
			minInd = ssk;
			break;
		elseif ssk == length(ssv)
			if iter > 1
				costmin = costTrue(iter-1);
			end
			minInd = ssk;
			if magGrad(iter) > gradDelta * gradEps
				status = 2;
				W=Wopts(:,:,:,ssk-2);
				Eflag = true;
				costmin = costopts(end);
				if BFGS == 1
					Bk = eye(2*N^2*K);
				end
			elseif sigmaChange(iter) < sigmaEps
				disp('decrease not possible.. but close enough..');
				W = W_old;
				status = 1;
			else
				disp('failed to converge');
				for k=1:K
					ww = W(:,:,k);
					W(:,:,k) = ww / sqrtm(ww'*ww);
				end
				status = 1;
			end
		end
	end
	
	sk = W - W_old;
	
	if magGrad(iter) == min(magGrad(1:iter))
		W_best = W;
	end
	
	
	costTrue(iter) = costmin;
	for n = 1:N
		costTrue(iter) = costTrue(iter) - gammaln(K) - log(beta_final(n)) ...
			+ K/beta_final(n) * log(2) ...
			+ log(2*pi^K) + gammaln(K/beta_final(n)) + ...
			log(abs(det(sigmab(:,:,n))))/2 - K/2 * log(etan(n)/4);
	end
	
	%% if progress made advance, otherwise, stay and shrink step size
	for k=1:K
		max_delta_w = max_delta_w + norm(W(:,:,k)-W_old(:,:,k),'fro');
	end % k
	max_delta_w = max_delta_w / K / N / N;
	W_old=W;
	
	cost(iter) = costopts(minInd);%fW2;
	%% output iteration info
	if iter > 1
		costChange = costTrue(iter) - costTrue(iter-1);
	else
		costChange = costTrue(iter);
	end
	gradAveNum = 1;
	if iter >= gradAveNum
		gradAve(iter) = sum(magGrad(iter-gradAveNum+1:iter)) / gradAveNum;
	else
		gradAve(iter) = magGrad(iter);
	end
	sigmaChange(iter) = norm(sigmab(:) - sigmab_old(:)) / 4 ;%/ K^2 / N;
	if verbose
		fprintf('%d:%d\t%.6f\t%.6f\t%d\t%.6f\t%.6f\t%.6f\n', ...
			BFGS,iter,costTrue(iter),magGrad(iter),minInd,costChange,gradAve(iter),sigmaChange(iter));
		if supplyA
			fprintf('\n Step %d: MagGrad: %.4f, W change: %f, SS: %f', iter, magGrad(iter)/N/K/N, max_delta_w,alpha0);
		else
			%fprintf('\n Step %d: MagGrad: %.4f, W change: %f, Cost: %f', iter, magGrad(iter)/N/K/N, max_delta_w,cost(iter));
		end
	end % verbose
	mdw(iter) = max_delta_w;
	a0(iter) = minInd;%ssv(minInd);
	
	%% Check the termination condition
	if  (gradAve(iter) < gradEps && sigmaChange(lastSigmaEval) < gradEps) || iter == maxIter || max_delta_w < eps
		if gradAve(iter) < gradEps && sigmaChange(lastSigmaEval) < gradEps
			fprintf('MagGrad sufficiently small.. %d iters\n',iter);
			status = 0;
		end
		if iter == maxIter
			fprintf('CMGGD max iteration.. \n');
			status = 3;
		end
		if max_delta_w < eps
			fprintf('CMGGD delta_w < eps.. %d iters\n',iter);
		end
		if alpha0==alphaMin
			fprintf('CMGGD min step size reached.. \n');
		end
		break;
	elseif max_delta_w > blowup || isnan(cost(iter))
		for k = 1:K
			W(:,:,k) = eye(N) + 0.1*randn(N);
		end
		if verbose
			fprintf('\n W blowup, restart with new initial value.');
		end
	end
	
end % iter

W = W_best;
%max_dw = mdw(1:iter);
magFinal = magGrad(1:iter);
gradAve = gradAve(1:iter);
sigmaChange = sigmaChange(1:iter);
a0=a0(1:iter);
%     magFinal = magGrad(iter)/N/K/N;
%     fprintf('mag grad final: %.6f\n',magGrad(iter));
%     figure;semilogy(1:iter,magGrad(1:iter),'b-o','LineWidth',2.4);ylabel('mag grad');
if verbose
	lw=2.4;
	figure;
	subplot(2,2,1);
	semilogy(1:iter,costTrue(1:iter),'b-o','LineWidth',lw);ylabel('True cost');
	ylim([min(costTrue(1:iter)) max(costTrue(1:iter))]);
	
	subplot(2,2,2);
	semilogy(1:iter,magGrad(1:iter),'b-o','LineWidth',lw);ylabel('mag grad');
	
	subplot(2,2,3);
	semilogy(1:iter,gradAve(1:iter),'b-o','LineWidth',lw);ylabel('ave mag grad');
	
	subplot(2,2,4);
	semilogy(1:iter,sigmaChange(1:iter),'b-o','LineWidth',lw);ylabel('sigma change');
end


if verbose
	fprintf('\n');
end

cost=costTrue(1:iter);
if outputISI
	isi=isi(1:iter);
end

% Now we resort the order of the components so that SCV that is most ill-conditioned as
% possible.
detSCV=zeros(1,N);
for n=1:N
	% Efficient version of Sigma_n=Yn*Yn'/T;
	Sigma_n2=eye(K); %
	for k1=1:K
		for k2=(1+k1):K
			Sigma_n2(k1,k2)=W(n,:,k1)*Rx{k1,k2}*W(n,:,k2)';
			Sigma_n2(k2,k1)=Sigma_n2(k1,k2)';
		end % k2
	end %k3
	detSCV(n)=det(Sigma_n2);
end

[~,isort]=sort(detSCV);

% Sort W for unwhitened data.
for k=1:K
	W(:,:,k)=W(isort,:,k);
end

if whiten_opt
	for k=1:K
		W(:,:,k)=W(:,:,k)*V(:,:,k);
		W_best(:,:,k) = W_best(:,:,k) * V(:,:,k);
	end
end

%% output var
res.W = W;
res.cost = cost;
res.isi = isi;
res.beta_final = beta_final;
res.iter = iter;
res.magFinal = magFinal;
res.gradAve = gradAve;
res.sigmaChange = sigmaChange;
res.status = status;
end

function [fW, dfW] = FunCost(W,X,sigmab,betavec,etan)
[N,T,K] = size(X);
fW = 0;
Y = zeros(N,T,K);
if nargout > 1
	dfW = zeros(N,N,K);
end
for k = 1:K
	Y(:,:,k) = W(:,:,k) * X(:,:,k);
end
for n=1:N
	if K == 1
		Yn = Y(n,:,1);
	else
		Yn = squeeze(Y(n,:,:)).';
	end
	Ynb = [real(Yn);imag(Yn)];
	sigmaInvYnb = sigmab(:,:,n)\Ynb;
	quadd = dot(Ynb,sigmaInvYnb);
	betan = betavec(n);
	fW = fW + mean(quadd.^betan) / 2 * etan(n)^betan;
	if nargout > 1
		dfW(n,:,:) = FunGrad(W(n,:,:), W([1:n-1 n+1:N],:,:), ...
			quadd.^(betan-1),[eye(K) 1j*eye(K); eye(K) -1j*eye(K)]*sigmaInvYnb, ...
			conj(X),etan(n),betan);
	end
end
for k = 1:K
	fW = fW - 2*log(abs(det(W(:,:,k))));
end
end

function dWn = FunGrad(Wn,Wnout,qbm1,sigmaInvYna,Xc,etan,betan)
[~,N,K] = size(Wn);
dWn = zeros(1,N,K);
for k=1:K
	[Q,~]=qr(Wnout(:,:,k)');
	hnk=Q(:,end);
	phi = qbm1 .* sigmaInvYna(k,:);
	g1 = mean(bsxfun(@times, Xc(:,:,k), phi),2);
	g2 = betan / 4 * etan^betan * g1;
	den = Wn(1,:,k)*hnk;
	gradNorm = conj(g2) - hnk/den;
	wnk=Wn(1,:,k)';
	gradProjMat = wnk * wnk';
	gradNormProj=(eye(N) - gradProjMat) * gradNorm;
	dWn(1,:,k) = gradNormProj';
end
end

function Wout = FunNorm(Win,normW)
[N,~,K] = size(Win);
Wout = Win;
for k=1:K
	if normW == 1
		for n=1:N
			Wout(n,:,k) = Win(n,:,k) ./ norm(Win(n,:,k));
		end
	else
		ww = Wout(:,:,k);
		Wout(:,:,k) = ww / sqrtm(ww'*ww);
	end
end
end

function [z,V,U]=whiten(x)
% [z,V,U]=whiten(x)
%
% Whitens the data vector so that E{zz'}=I, where z=V*x.

if ~iscell(x)
	[N,T,K]=size(x);
	if K==1
		% Step 1. Center the data.
		x=bsxfun(@minus,x,mean(x,2));
		
		% Step 2. Form MLE of data covariance.
		covar=x*x'/T;
		
		% Step 3. Eigen decomposition of covariance.
		[eigvec, eigval] = eig (covar);
		
		% Step 4. Forming whitening transformation.
		V=sqrt(eigval) \ eigvec';
		U=eigvec * sqrt(eigval);
		
		% Step 5. Form whitened data
		z=V*x;
	else
		K=size(x,3);
		z=zeros(N,T,K);
		V=zeros(N,N,K);
		U=zeros(N,N,K);
		for k=1:K
			% Step 1. Center the data.
			xk=bsxfun(@minus,x(:,:,k),mean(x(:,:,k),2));
			
			% Step 2. Form MLE of data covariance.
			covar=xk*xk'/T;
			
			% Step 3. Eigen decomposition of covariance.
			[eigvec, eigval] = eig (covar);
			
			% Step 4. Forming whitening transformation.
			V(:,:,k)=sqrt(eigval) \ eigvec';
			U(:,:,k)=eigvec * sqrt(eigval);
			
			% Step 5. Form whitened data
			z(:,:,k)=V(:,:,k)*xk;
		end % k
	end % K>1
else % x is cell
	K=numel(x);
	sizex=size(x);
	V=cell(sizex);
	U=cell(sizex);
	z=cell(sizex);
	for k=1:K
		T=size(x{k},2);
		% Step 1. Center the data.
		xk=bsxfun(@minus,x{k},mean(x{k},2));
		
		% Step 2. Form MLE of data covariance.
		covar=xk*xk'/T;
		
		% Step 3. Eigen decomposition of covariance.
		[eigvec, eigval] = eig (covar);
		
		% Step 4. Forming whitening transformation.
		V{k}=sqrt(eigval) \ eigvec';
		U{k}=eigvec * sqrt(eigval);
		
		% Step 5. Form whitened data
		z{k}=V{k}*xk;
	end % k
end

end